#include "Socket.h"

Socket::Socket(Tango::DeviceImpl* devImpl, const std::string host_name, const int port_number, const long timeout_msec, const bool try_recon) :
        Tango::LogAdapter(devImpl),  hostname(host_name), portN(port_number), try_reconnection(try_recon)
{
    DEBUG_STREAM << "Socket::Socket() " << printTime() << endl;

    timeout.tv_sec = int(timeout_msec/1000);
    DEBUG_STREAM << "Socket::Socket() timeout sec: " << timeout.tv_sec << endl;
    timeout.tv_usec = (timeout_msec % 1000) * 1000;
    DEBUG_STREAM << "Socket::Socket() timeout usec: " << timeout.tv_usec << endl;
}

Socket::~Socket()
{
    DEBUG_STREAM << "Socket::~Socket() " << printTime() << endl;

    if(socketfd != -1)
    {
        DEBUG_STREAM << "Socket::~Socket() deleting socket file descriptor... " << endl;
        close(socketfd);
        socketfd = -1;
    }
}

void Socket::open_socket()
{
    DEBUG_STREAM << "Socket::open_socket() " << printTime() << endl;
    
    DEBUG_STREAM << "Socket::open_socket() hostname: " << hostname << " port number: " << portN << endl;

    char ipstr[INET6_ADDRSTRLEN];

    struct sockaddr_in sa;
    memset(&sa, 0, sizeof(sa));

    struct sockaddr_in6 sa6;
    memset(&sa6, 0, sizeof(sa6));

    struct addrinfo hints;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    struct addrinfo *res, *p;

    if( inet_pton(AF_INET, hostname.c_str(), &(sa.sin_addr))>0 )
    {
        DEBUG_STREAM << "Socket::open_socket() " << hostname << " IPv4"<< endl;
        sa.sin_family = AF_INET;
        sa.sin_port = htons(portN);
        socketfd = socket(PF_INET, SOCK_STREAM, 0);
        connect_socket((struct sockaddr*) &sa);
    }
    else if( inet_pton(AF_INET6, hostname.c_str(), &(sa6.sin6_addr))>0 )
    {
        DEBUG_STREAM << "Socket::open_socket() " << hostname << " IPv6"<< endl;
        sa6.sin6_family = AF_INET6;
        sa6.sin6_port = htons(portN);
        socketfd = socket(PF_INET6, SOCK_STREAM, 0);
        connect_socket((struct sockaddr*) &sa6);
    }
    else if( getaddrinfo(hostname.c_str(), NULL, &hints, &res)==0 )
    {
        DEBUG_STREAM << "Socket::open_socket() DNS lookup for " << hostname << endl;
        for(p=res; p!=NULL; p=p->ai_next)
        {
            if(p->ai_family == AF_INET)
            {
                DEBUG_STREAM << "Socket::open_socket() " << hostname << " IPv4"<< endl;
                inet_ntop(p->ai_family, &(((struct sockaddr_in *)p->ai_addr)->sin_addr), ipstr, sizeof ipstr);
                DEBUG_STREAM << "Socket::open_socket() address: " << ipstr << endl;
                sa.sin_addr = ((struct sockaddr_in *)p->ai_addr)->sin_addr;
                sa.sin_family = ((struct sockaddr_in *)p->ai_addr)->sin_family;
                sa.sin_port = htons(portN);
                socketfd = socket(PF_INET, SOCK_STREAM, 0);
                connect_socket((struct sockaddr*) &sa);
                if(socketfd != -1)
                    break;
            }
            else
            {
                DEBUG_STREAM << "Socket::open_socket() " << hostname << " IPv6"<< endl;
                inet_ntop(p->ai_family, &(((struct sockaddr_in6 *)p->ai_addr)->sin6_addr), ipstr, sizeof ipstr);
                DEBUG_STREAM << "Socket::open_socket() address: " << ipstr << endl;
                sa6.sin6_addr = ((struct sockaddr_in6 *)p->ai_addr)->sin6_addr;
                sa6.sin6_family = ((struct sockaddr_in6 *)p->ai_addr)->sin6_family;
                sa6.sin6_port = htons(portN);
                socketfd = socket(PF_INET6, SOCK_STREAM, 0);
                connect_socket((struct sockaddr*) &sa6);
                if(socketfd != -1)
                    break;
            }
        }        
        freeaddrinfo(res);
    }
    else
    {
        WARN_STREAM << "Socket::open_socket() Unkown hostname " << printTime() << endl;
        throw SocketException("Unkown hostname");
    }
    
    if(socketfd <= 0)
    {
        WARN_STREAM << "Socket::open_socket() error creating new socket " << printTime() << endl;
        throw SocketException("Error creating new socket");
    }
    else
    {
        DEBUG_STREAM << "Socket::open_socket() connection established " << printTime() << endl;
        omni_thread::sleep(1,0);
        char buffer[1000];
        read(buffer, 1000);
        DEBUG_STREAM << "Socket::open_socket() socket intial data cleanded up " << printTime() << endl;
    }
}

void Socket::connect_socket(const struct sockaddr* s_addr)
{
    DEBUG_STREAM << "Socket::create_socket() " << printTime() << endl;

    int arg, res, valopt;
    fd_set fdset;
    socklen_t lon;
    
    if(socketfd == -1)
    {
        return;
    }
    if((arg=fcntl(socketfd, F_GETFL, NULL)) == -1)
    {
        WARN_STREAM << "Socket::create_socket() fcntl(...,F_GETFL) " << strerror(errno) << " " << printTime() << endl;
        close_socket();
        return;
    }
    arg |= O_NONBLOCK;
    if(fcntl(socketfd, F_SETFL, arg) == -1)
    {
        WARN_STREAM << "Socket::create_socket() fcntl(...,F_SETFL) " << strerror(errno) << " " << printTime() << endl;
        close_socket();
        return;
    }
    res = connect(socketfd, s_addr, sizeof(struct sockaddr));
    if(res<0)
    {
        if(errno == EINPROGRESS)
        {
            DEBUG_STREAM << "Socket::create_socket() connection in progress " << printTime() << endl;
            do
            {
                FD_ZERO(&fdset);
                FD_SET(socketfd, &fdset);
                timeval tv = timeout;
                res = select(socketfd+1, NULL, &fdset, NULL, &tv);
                if(res < 0 && errno != EINTR)
                {
                    WARN_STREAM << "Socket::create_socket() error on select " << strerror(errno) << " " << printTime() << endl;
                    close_socket();
                    return;
                }
                else if(res > 0)
                {
                    lon = sizeof(int);
                    if(getsockopt(socketfd,  SOL_SOCKET, SO_ERROR, (void*)(&valopt), &lon) < 0)
                    {
                        WARN_STREAM << "Socket::create_socket() error on getsockopt " << strerror(errno) << " " << printTime() << endl;
                        close_socket();
                        return;
                    }
                    if(valopt)
                    {
                        WARN_STREAM << "Socket::create_socket() error on delayed connection " << strerror(valopt) << " " << printTime() << endl;
                        close_socket();
                        return;
                    }                    
                    break;
                }
            }
            while(1);
        }
        else
        {
            WARN_STREAM << "Socket::create_socket() error connecting " << strerror(errno) << " " << printTime() << endl;
            close_socket();
            return;
        }
    }
}

void Socket::close_socket()
{
    DEBUG_STREAM << "Socket::close_socket() " << printTime() << endl;
    
    if(socketfd != -1)
    {
        close(socketfd);
        socketfd = -1;
    }
}

int Socket::write(const char* buffer, const int buff_size)
{
    DEBUG_STREAM << "Socket::write() start... " << printTime() << endl;

    int res;
    fd_set fdset;
    if(socketfd <= 0)
    {
        if(try_reconnection)
        {
            WARN_STREAM << "Socket::write() try to reopen connection " << printTime() << endl;
            open_socket();
        }
        if(socketfd <=0)
        {
            WARN_STREAM << "Socket::write() fail on reopening connection " << printTime() << endl;
            throw SocketException("fail on reopening connection");
        }
    }
    FD_ZERO(&fdset);
    FD_SET(socketfd, &fdset);
    timeval tv = timeout;
    res = select(socketfd+1, NULL, &fdset, NULL, &tv);
    if(res < 0)
    {
        WARN_STREAM << "Socket::write() error on select " << strerror(errno) << " " << printTime() << endl;
        close_socket();
        throw SocketException("Error error on select");
    }
    else if(res == 0)
    {
        DEBUG_STREAM << "Socket::read() timeout" << endl;
        return 0;
    }
    if(FD_ISSET(socketfd, &fdset))
    {
        FD_CLR(socketfd, &fdset);
        for(int i=0; i<buff_size; i++)
        {
            DEBUG_STREAM << "| 0x" << hex << int(buffer[i]) << " | " << buffer[i] << endl;
        }
        res = send(socketfd, buffer, buff_size, 0);
        if(res <= 0)
        {
            WARN_STREAM << "Socket::write() error on send " << strerror(errno) << " " << printTime() << endl;
            close_socket();
            throw SocketException("Error error on send");
        }
    }
    DEBUG_STREAM << "Socket::write() stop... " << printTime() << endl;
    return res;
}

int Socket::read(char* buffer, const int buff_size)
{
    DEBUG_STREAM << "Socket::read() start... " << printTime() << endl;

    int res;
    fd_set fdset;
    if(socketfd <= 0)
    {
        if(try_reconnection)
        {
            WARN_STREAM << "Socket::read() try to reopen connection " << printTime() << endl;
            open_socket();
        }
        if(socketfd <=0)
        {
            WARN_STREAM << "Socket::read() fail on reopening connection " << printTime() << endl;
            throw SocketException("fail on reopening connection");
        }
    }
    FD_ZERO(&fdset);
    FD_SET(socketfd, &fdset);
    timeval tv = timeout;
    res = select(socketfd+1, &fdset,  NULL, NULL, &tv);
    if(res < 0)
    {
        WARN_STREAM << "Socket::read() error on select " << strerror(errno) << " " << printTime() << endl;
        close_socket();
        throw SocketException("Error error on select");
    }
    else if(res == 0)
    {
        DEBUG_STREAM << "Socket::read() buffer empty, timeout " << printTime() << endl;
        return 0;
    }
    if(FD_ISSET(socketfd, &fdset))
    {
        FD_CLR(socketfd, &fdset);
        res = recv(socketfd, buffer, buff_size, 0);
        if(res < 0)
        {
            WARN_STREAM << "Socket::read() error on recv " << strerror(errno) << " " << printTime() << endl;
            close_socket();
            throw SocketException("Error error on recv");
        }
        for(int i=0; i<res; i++)
        {
            DEBUG_STREAM << "| 0x" << hex << int(buffer[i]) << " | " << buffer[i] << endl;
        }
    }
    DEBUG_STREAM << "Socket::read() stop... " << printTime() << endl;
    return res;
}

long Socket::left_to_read()
{
    DEBUG_STREAM << "Socket::leftToRead() " << printTime() << endl;
    
    long to_read = 0;
    if(socketfd <= 0)
    {
        if(try_reconnection)
        {
            WARN_STREAM << "Socket::read() try to reopen connection " << printTime() << endl;
            open_socket();
            return 0;
        }
        if(socketfd <=0)
        {
            WARN_STREAM << "Socket::read() fail on reopening connection " << printTime() << endl;
            throw SocketException("fail on reopening connection");
        }
    }
    if(ioctl(socketfd, FIONREAD, &to_read) < 0)
    {
        WARN_STREAM << "Socket::left_to_read() error on ioctl " << strerror(errno) << " " << printTime() << endl;
        close_socket();
        throw SocketException("Error error on ioctl");
    }    
    DEBUG_STREAM << "Socket::leftToRead() " << to_read << " on buffer " << printTime() << endl;
    return to_read;
}

std::string Socket::printTime()
{
    std::stringstream s;
    struct timeval tv;
    gettimeofday(&tv, NULL);

    int days = tv.tv_sec/86400;
    int sec_of_day = tv.tv_sec - (days*86400);
    int hour = sec_of_day/3600;
    int min_left = sec_of_day-(hour*3600);
    int min = min_left/60;
    int sec = min_left-(min*60);
    int msec = tv.tv_usec/1000;

    s << "@ " << hour+1 << ":" << min << ":" << sec << ":" << msec << endl;
    return s.str();
}
